#include <string.h>

// typeless representation of a function or object/mem-func

class FunctorBase
{
public:
	typedef void (FunctorBase::*_MemFunc)();
	typedef void (*_Func)();
	FunctorBase():callee(0),func(0) {;}
	FunctorBase(const void *c, const void *f, unsigned sz)
	{
		if(c)	//must be callee/memfunc
		{
			callee = (void *)c;
			memcpy(memFunc,f,sz);
		}
		else	//must be ptr-to-func
		{
			func = f;
		}
	}
	// for evaluation in conditions
	// will be changed to bool when bool exists
	operator int() const { return func||callee; }

	class DummyInit
	{
	};
////////////////////////////////////////////////////////////////
// Note: this code depends on all ptr-to-mem-funcs being same size
// If that is not the case then make memFunc as large as largest
////////////////////////////////////////////////////////////////

	union
	{
		const void *func;
		char memFunc[sizeof(_MemFunc)];
	};
	void *callee;
};

// The Functor class is a template. It has a default constructor and the
// required operator() corresponding to its template parameters. It uses the
// generated copy constructor and assignment operators.

/************************* one arg - no return *******************/
/*
template <class P1> class Functor1 : protected FunctorBase
{
public:
	Functor1(DummyInit * = 0) {;}
	void operator()(P1 p1) const
	{
		thunk(*this,p1);
	}
	FunctorBase::operator int;
protected:
	typedef void (*Thunk)(const FunctorBase &, P1);
	Functor1(Thunk t, const void *c, const void *f, size_t sz):
		FunctorBase(c,f,sz),thunk(t) {;}
private:
	Thunk thunk;
};

template <class P1, class P2> class Functor2:protected FunctorBase
{
public:
	Functor2(DummyInit * = 0) {;}
	void operator()(P1 p1, P1 p2) const
	{
		thunk(*this,p1,p2);
	}
	FunctorBase::operator int;
protected:
	typedef void (*Thunk)(const FunctorBase &, P1, P2);
	Functor2(Thunk t,const void *c,const void *f, size_t sz):
		FunctorBase(c,f,sz),thunk(t) {;}
private:
	Thunk thunk;
};

template <class P1, class P2, class P3> class Functor3:protected FunctorBase
{
public:
	Functor3(DummyInit * = 0) {;}
	void operator()(P1 p1, P1 p2, P3 p3) const
	{
		thunk(*this,p1,p2,p3);
	}
	FunctorBase::operator int;
protected:
	typedef void (*Thunk)(const FunctorBase &, P1, P2, P3);
	Functor3(Thunk t,const void *c,const void *f, size_t sz):
		FunctorBase(c,f,sz),thunk(t) {;}
private:
	Thunk thunk;
};
*/
template <class RT, class P1>
	class Functor1 : protected FunctorBase
{
public:
	Functor1(DummyInit * = 0) {;}
	RT operator()(P1 p1) const
	{
		return thunk(*this, p1);
	}
	FunctorBase::operator int;
protected:
	typedef RT (*Thunk)(const FunctorBase &, P1);
	Functor1(Thunk t, const void *c, const void *f, size_t sz):
		FunctorBase(c,f,sz),thunk(t) { }
private:
	Thunk thunk;
};

template <class RT, class P1, class P2>
class Functor2: protected FunctorBase
{
public:
	Functor2(DummyInit * = 0) {;}
	RT operator()(P1 p1, P2 p2) const
	{
		return thunk(*this, p1, p2);
	}
	FunctorBase::operator int;
protected:
	typedef RT (*Thunk)(const FunctorBase &, P1, P2);
	Functor2(Thunk t, const void *c, const void *f, size_t sz):
		FunctorBase(c,f,sz),thunk(t) {;}
private:
	Thunk thunk;
};

template <class RT, class P1, class P2, class P3>
class Functor3: protected FunctorBase
{
public:
	Functor3(DummyInit * = 0) {;}
	RT operator()(P1 p1, P2 p2, P3 p3) const
	{
		return thunk(*this, p1, p2, p3);
	}
	FunctorBase::operator int;
protected:
	typedef RT (*Thunk)(const FunctorBase &, P1, P2, P3);
	Functor3(Thunk t, const void *c, const void *f, size_t sz):
		FunctorBase(c,f,sz),thunk(t) {;}
private:
	Thunk thunk;
};

/*
 The Functor class has a protected constructor that takes the same typeless args
 as FunctorBase, plus an additional first argument. This argument is a pointer
 to function (the thunk function) that takes the same arguments as the operator(),
 plus an additional first argument of type const FunctorBase &. The Functor stores
 this away (in thunk) and implements operator() by calling thunk(), passing itself
 and the other arguments. Thus it is this thunk() function that does the work of
 'calling back'.

 A key issue at this point is whether operator() should be virtual. In the first
 iteration of my mechanism the Functor classes were abstract and the operator()'s
 pure virtual. To use them for callbacks a set of derived template classes
 parameterized on the callee type was provided. This required that functors always
 be passed and held by reference or pointer and never by value. It also required
 the caller component or the client code maintain the derived object for as long
 as the callback relationship existed. I found the maintenance and lifetime issues
 of these functor objects to be problematic, and desired by-value syntax.

 In the current mechanism the Functor classes are concrete and the operator()
 is non-virtual. They can be treated and used just like ptr-to-functions. In
 particular, they can be stored by value in the component classes.

 The Translators
-------------------
Where does the thunk() come from? It is generated by the compiler as a static
member of a template 'translator' class. For each Functor class there are two
translator classes, one for stand-alone functions (FunctionTranslator) and one
for member functions (MemberTranslator). The translator classes are
parameterized by the type of the Functor as well as the type(s) of the callee.
With this knowledge they can, in a fully type-safe manner, perform two
important tasks.

First, they can initialize the Functor data. They do this by being publicly
derived from the Functor. They are constructed with typed callee information
and which they pass (untyped) to the functor's protected constructor.

Second, they have a static member function thunk(), which, when passed a
FunctorBase, converts its callee data back into typed information, and
executes the callback on the callee. It is a pointer to this static function
which is passed to the Functor constructor.
*/

template <class RT, class P1, class Func>
	class FunctionTranslator1 : public Functor1<RT, P1>
{
public:
	FunctionTranslator1(Func f): Functor1<RT, P1>(thunk, 0, f, 0) {;}
	static RT thunk(const FunctorBase &ftor, P1 p1)
	{
		return (Func(ftor.func))(p1);
	}
};

/*
FunctionTranslator is the simpler of the two. It is parameterized by the
argument type of the Functor and some ptr-to-function type (Func).
Its constructor takes an argument of type Func and passes it and a pointer to
its static thunk() function to the base class constructor. The thunk function,
given a FunctorBase ftor, casts ftor's func member back to its correct type
(Func) and calls it. There is an assumption here that the FunctorBase ftor is
one initialized by the constructor (or a copy). There is no danger of it being
otherwise, since the functors are always initialized with matching callee data
and thunk functions. This is what is called a 'safe' cast, since the same
entity that removed the type information also re-instates it, and can guarantee
a match. If Func's signature is incompatible with the call, i.e. if it cannot
be called with a single argument of type P1, then thunk() will not compile. If
implicit conversions are required the compiler will perform them.
*/

template <class RT, class P1,class Callee, class MemFunc>
	class MemberTranslator1:public Functor1<RT, P1>
{
public:
	MemberTranslator1(Callee &c,const MemFunc &m):
		Functor1<RT, P1>(thunk,&c,&m,sizeof(MemFunc)) {;}
	static RT thunk(const FunctorBase &ftor, P1 p1)
	{
		Callee *callee = (Callee *)ftor.callee;
		MemFunc &memFunc(*(MemFunc*)(void *)(ftor.memFunc));
		return (callee->*memFunc)(p1);
	}
};

/*
MemberTranslator is parameterized by the argument type of the Functor, some
class type (Callee), and some ptr-to-member-function type (MemFunc).
Not surprisingly it's constructor is passed 2 arguments, a Callee object
(by reference) and a ptr-to-member-function, both of which are passed,
along with the thunk function, to the base class constructor. Once again,
the thunk function casts the typeless info back to life, and then calls
the member function on the object, with the passed parameter.

Since the Translator objects are Functor objects, and fully 'bound' ones
at that, they are suitable initializers for their corresponding Functor,
using the Functor's copy constructor. We needn't worry about the 'chopping'
effect since the data is all in the base class portion of the Translator
class and there are no virtual functions involved. Thus they are perfect
candidates for the return value of makeFunctor()!

The makeFunctor Functions
--------------------------
For each Functor class there are three versions of makeFunctor(), one for
ptr-to-function and a const and non-const version for the
object/ptr-to-member-function pair.
*/

template <class P1, class TRT, class TP1>
	inline FunctionTranslator1<TRT, P1,TRT (*)(TP1)>
		makeFunctor(Functor1<TRT, P1>*, TRT (*f)(TP1))
{
	return FunctionTranslator1<TRT, P1, TRT (*)(TP1)>(f);
}

/*
The function version is straightforward. It uses the dummy argument to tell
it the type of the functor and merely returns a corresponding FunctionTranslator.
I mentioned above that the Func type parameter of FunctionTranslator was
invariably a ptr-to-function type. This version of makeFunctor() ensures that
by explicity specifying it as such.
*/

template <class P1,class Callee,class TRT,class CallType,class TP1>
	inline MemberTranslator1<TRT, P1, Callee, TRT (CallType::*)(TP1)>
		makeFunctor(Functor1<TRT, P1>*,Callee &c,TRT (CallType::* const &f)(TP1))
{
	typedef TRT (CallType::*MemFunc)(TP1);
	return MemberTranslator1<TRT, P1, Callee, MemFunc>(c,f);
}

/*
This is the gnarliest bit. Here makeFunctor is parameterized with the type
of the argument to the Functor, the type of the callee, the type of the
class of which the member-function is a member, the argument and return
types of the member function. Whew! We're a long way from Stack<T> land!
Like the ptr-to-function version, it uses the dummy first argument of the
constructor to determine the type of the Functor. The second argument is
a Callee object (by reference). The third argument is this thing:

TRT (CallType::* const &f)(TP1)

Here f is a reference to a constant pointer to a member function of
CallType taking TP1 and returning TRT. You might notice that
pointer-to-member-functions are all handled by reference in the library.
On some implementations they can be expensive to pass by value and copy.
The significant feature here is that the function need not be of type
pointer-to-member-of-Callee. This allows makeFunctor to match on (and
ultimately work with) a ptr-to-member-function of some base of Callee.
It then typedefs that bit and returns an appropriate MemberTranslator.
*/

template <class P1,class Callee,class TRT,class CallType,class TP1>
	inline MemberTranslator1<TRT,P1,const Callee,TRT (CallType::*)(TP1)const>
		makeFunctor(TRT,Functor1<TRT, P1>*,const Callee &c,TRT (CallType::* const &f)(TP1)const)
{
	typedef TRT (CallType::*MemFunc)(TP1)const;
	return MemberTranslator1<TRT, P1,const Callee,MemFunc>(c,f);
}

